# -*- coding: utf-8 -*-
#

# Imports
from .LogisticMapDataset import LogisticMapDataset
from .MackeyGlassDataset import MackeyGlassDataset
from .MemTestDataset import MemTestDataset
from .NARMADataset import NARMADataset

__all__ = [
    'LogisticMapDataset', 'MackeyGlassDataset', 'MemTestDataset', 'NARMADataset'
]



# -*- coding: utf-8 -*-
#

# Imports
import torch
from torch.utils.data.dataset import Dataset


# 10th order NARMA task
class NARMADataset(Dataset):
    """
    xth order NARMA task
    WARNING: this is an unstable dataset. There is a small chance the system becomes
    unstable, leading to an unusable dataset. It is better to use NARMA30 which
    where this problem happens less often.
    """

    # Constructor
    def __init__(self, sample_len, n_samples, system_order=10, seed=None):
        """
        Constructor
        :param sample_len: Length of the time-series in time steps.
        :param n_samples: Number of samples to generate.
        :param system_order: th order NARMA
        :param seed: Seed of random number generator.
        """
        # Properties
        self.sample_len = sample_len
        self.n_samples = n_samples
        self.system_order = system_order

        # System order
        self.parameters = torch.zeros(4)
        if system_order == 10:
            self.parameters[0] = 0.3
            self.parameters[1] = 0.05
            self.parameters[2] = 9
            self.parameters[3] = 0.1
        else:
            self.parameters[0] = 0.2
            self.parameters[1] = 0.04
            self.parameters[2] = 29
            self.parameters[3] = 0.001
        # end if

        # Init seed if needed
        if seed is not None:
            torch.manual_seed(seed)
        # end if

        # Generate data set
        self.inputs, self.outputs = self._generate()
    # end __init__

    #############################################
    # OVERRIDE
    #############################################

    # Length
    def __len__(self):
        """
        Length
        :return:
        """
        return self.n_samples
    # end __len__

    # Get item
    def __getitem__(self, idx):
        """
        Get item
        :param idx:
        :return:
        """
        return self.inputs[idx], self.outputs[idx]
    # end __getitem__

    ##############################################
    # PRIVATE
    ##############################################

    # Generate
    def _generate(self):
        """
        Generate dataset
        :return:
        """
        inputs = list()
        outputs = list()
        for i in range(self.n_samples):
            ins = torch.rand(self.sample_len, 1) * 0.5
            outs = torch.zeros(self.sample_len, 1)
            for k in range(self.system_order - 1, self.sample_len - 1):
                outs[k + 1] = self.parameters[0] * outs[k] + self.parameters[1] * outs[k] * torch.sum(
                    outs[k - (self.system_order - 1):k + 1]) + 1.5 * ins[k - int(self.parameters[2])] * ins[k] + \
                                 self.parameters[3]
            # end for
            inputs.append(ins)
            outputs.append(outs)
        # end for

        return inputs, outputs
    # end _generate

# end NARMADataset



# -*- coding: utf-8 -*-
#
# File : echotorch/nn/ESN.py
# Description : An Echo State Network module.
# Date : 26th of January, 2018
#
# This file is part of EchoTorch.  EchoTorch is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Nils Schaetti <nils.schaetti@unine.ch>

"""
Created on 26 January 2018
@author: Nils Schaetti
"""

# Imports
import torch.sparse
import torch
import torch.nn as nn
import numpy as np
from past.utils import old_div


# Slow Feature Analysis layer
class SFACell(nn.Module):
    """
    Extract the slowly varying components from input data.
    """

    # Type keys
    _type_keys = ['f', 'd', 'F', 'D']

    # Type conv
    _type_conv = {('f', 'd'): 'd', ('f', 'F'): 'F', ('f', 'D'): 'D',
                  ('d', 'F'): 'D', ('d', 'D'): 'D',
                  ('F', 'd'): 'D', ('F', 'D'): 'D'}

    # Constructor
    def __init__(self, input_dim, output_dim, include_last_sample=True, rank_deficit_method='none', use_bias=True):
        """
        Constructor
        :param input_dim: Input dimension
        :param output_dim: Number of slow feature
        :param include_last_sample: If set to False, the training method discards the last sample in every chunk during training when calculating the matrix.
        :param rank_deficit_method: 'none', 'reg', 'pca', 'svd', 'auto'.
        """
        super(SFACell, self).__init__()
        self.include_last_sample = include_last_sample
        self.use_bias = use_bias
        self.input_dim = input_dim
        self.output_dim = output_dim

        # Initialie the two covariance matrices one for
        # the input data, and the other for the derivatives.
        self.xTx = torch.zeros(input_dim, input_dim)
        self.xTx_avg = torch.zeros(input_dim)
        self.dxTdx = torch.zeros(input_dim, input_dim)
        self.dxTdx_avg = torch.zeros(input_dim)

        # Set routine for eigenproblem
        self.set_rank_deficit_method(rank_deficit_method)
        self.rank_threshold = 1e-12
        self.rank_deficit = 0

        # Will be set after training
        self.d = None
        self.sf = None
        self.avg = None
        self.bias = None
        self.tlen = None
    # end __init__

    ###############################################
    # PROPERTIES
    ###############################################

    ###############################################
    # PUBLIC
    ###############################################

    # Time derivative
    def time_derivative(self, x):
        """
        Compute the approximation of time derivative
        :param x:
        :return:
        """
        return x[1:, :] - x[:-1, :]
    # end time_derivative

    # Reset learning
    def reset(self):
        """
        Reset learning
        :return:
        """
        # Training mode again
        self.train(True)
    # end reset

    # Forward
    def forward(self, x):
        """
        Forward
        :param x: Input signal.
        :return: Output or hidden states
        """
        # For each batch
        for b in np.arange(0, x.size(0)):
            # If training or execution
            if self.training:
                # Last sample
                last_sample_index = None if self.include_last_sample else -1

                # Sample and derivative
                xs = x[b, :last_sample_index, :]
                xd = self.time_derivative(x[b])

                # Update covariance matrix
                self.xTx.data.add(xs.t().mm(xs))
                self.dxTdx.data.add(xd.t().mm(xd))

                # Update average
                self.xTx_avg += torch.sum(xs, axis=1)
                self.dxTdx_avg += torch.sum(xd, axis=1)

                # Length
                self.tlen += x.size(0)
            else:
                x[b].mv(self.sf) - self.bias
            # end if
        # end if
        return x
    # end forward

    # Finish training
    def finalize(self):
        """
        Finalize training with LU factorization or Pseudo-inverse
        """
        # Covariance
        self.xTx, self.xTx_avg, self.tlen = self._fix(self.xtX, self.xTx_avg, self.tlen, center=True)
        self.dxTdx, self.dxTdx_avg, self.tlen = self._fix(self.dxTdx, self.dxTdx_avg, self.tlen, center=False)

        # Range
        rng = (1, self.output_dim)

        # Resolve system
        self.d, self.sf = self._symeig(
            self.dxTdx, self.xTx, rng
        )
        d = self.d

        # We want only positive values
        if torch.min(d) < 0:
            raise Exception(u"Got negative values in {}".format(d))
        # end if

        # Delete covariance matrix
        del self.xTx
        del self.dxTdx

        # Store bias
        self.bias = self.xTx_avg * self.sf
    # end finalize

    ###############################################
    # PRIVATE
    ###############################################

    # Solve standard and generalized eigenvalue problem for symmetric (hermitian) definite positive matrices
    def _symeig(self, A, B, range, eigenvectors=True):
        """
        Solve standard and generalized eigenvalue problem for symmetric (hermitian) definite positive matrices.
        :param A: An N x N matrix
        :param B: An N x N matrix
        :param range: (lo, hi), the indexes of smallest and largest eigenvalues to be returned.
        :param eigenvectors: Return eigenvalues and eigenvector or only engeivalues
        :return: w, the eigenvalues and Z the eigenvectors
        """
        # To numpy
        A = A.numpy()
        B = B.numpy()

        # Type
        dtype = np.dtype()

        # Make B the identity matrix
        wB, ZB = np.linalg.eigh(B)

        # Check eigenvalues
        self._assert_eigenvalues_real(wB)

        # No negative values
        if wB.real.min() < 0:
            raise Exception(u"Got negative eigenvalues: {}".format(wB))
        # end if

        # Old division
        ZB = old_div(ZB.real, np.sqrt(wB.real))

        # A = ZB^T * A * ZB
        A = np.matmul(np.matmul(ZB.T, A), ZB)

        # Diagonalize A
        w, ZA = np.linalg.eigh(A)
        Z = np.matmul(ZB, ZA)

        # Check eigenvalues
        self._assert_eigenvalues_real(w, dtype)

        # Read
        w = w.real
        Z = Z.real

        # Sort
        idx = w.argsort()
        w = w.take(idx)
        Z = Z.take(idx, axis=1)

        # Sanitize range
        n = A.shape[0]
        lo, hi = range
        if lo < 1:
            lo = 1
        # end if
        if lo > n:
            lo = n
        # end if
        if hi > n:
            hi = n
        # end if
        if lo > hi:
            lo, hi = hi, lo
        # end if

        # Get values
        Z = Z[:, lo-1:hi]
        w = w[lo-1:hi]

        # Cast
        w = self.refcast(w, dtype)
        Z = self.refcast(Z, dtype)

        # Eigenvectors
        if eigenvectors:
            return torch.FloatTensor(w), torch.FloatTensor(Z)
        else:
            return torch.FloatTensor(w)
        # end if
    # end _symeig

    # Ref cast
    def refcast(self, array, dtype):
        """
        Cast the array to dtype only if necessary, otherwise return a reference.
        """
        dtype = np.dtype(dtype)
        if array.dtype == dtype:
            return array
        return array.astype(dtype)
    # end refcast

    # Check eigenvalues
    def _assert_eigenvalues_real(self, w, dtype):
        """
        Check eigenvalues
        :param w:
        :param dtype:
        :return:
        """
        tol = np.finfo(dtype.type).eps * 100
        if abs(w.imag).max() > tol:
            err = "Some eigenvalues have significant imaginary part: %s " % str(w)
            raise Exception(err)
        # end if
    # end _assert_eigenvalues_real

    # Greatest common type
    def _greatest_common_dtype(self, alist):
        """
        Apply conversion rules to find the common conversion type
        dtype 'd' is default for 'i' or unknown types
        (known types: 'f','d','F','D').
        """
        dtype = 'f'
        for array in alist:
            if array is None:
                continue
            tc = array.dtype.char
            if tc not in self._type_keys:
                tc = 'd'
            transition = (dtype, tc)
            if transition in self._type_conv:
                dtype = self._type_conv[transition]
        return dtype
    # end _greatest_common_dtype

    # Fix covariance matrix
    def _fix(self, mtx, avg, tlen, center=True):
        """
        Returns a triple containing the covariance matrix, the average and
        the number of observations.
        :param mtx:
        :param center:
        :return:
        """
        if self.use_bias:
            mtx /= tlen
        else:
            mtx /= tlen - 1
        # end if

        # Substract the mean
        if center:
            avg_mtx = np.outer(avg, avg)
            if self.use_bias:
                avg_mtx /= tlen * tlen
            else:
                avg_mtx /= tlen * (tlen - 1)
            # end if
            mtx -= avg_mtx
        # end if

        # Fix the average
        avg /= tlen

        return mtx, avg, tlen
    # end fix

# end SFACell



# -*- coding: utf-8 -*-
#
# File : echotorch/nn/LiESNCell.py
# Description : An Leaky-Integrated Echo State Network layer.
# Date : 26th of January, 2018
#
# This file is part of EchoTorch.  EchoTorch is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Nils Schaetti <nils.schaetti@unine.ch>

"""
Created on 26 January 2018
@author: Nils Schaetti
"""

import torch
import torch.sparse
import torch.nn as nn
from torch.autograd import Variable
from .ESNCell import ESNCell
import matplotlib.pyplot as plt


# Leak-Integrated Echo State Network layer
class LiESNCell(ESNCell):
    """
    Leaky-Integrated Echo State Network layer
    """

    # Constructor
    def __init__(self, leaky_rate=1.0, train_leaky_rate=False, *args, **kwargs):
        """
        Constructor
        :param leaky_rate: Reservoir's leaky rate (default 1.0, normal ESN)
        :param train_leaky_rate: Train leaky rate as parameter? (default: False)
        """
        super(LiESNCell, self).__init__(*args, **kwargs)

        # Params
        if train_leaky_rate:
            self.leaky_rate = nn.Parameter(torch.Tensor(1).fill_(leaky_rate), requires_grad=True)
        else:
            # Initialize bias
            self.register_buffer('leaky_rate', Variable(torch.Tensor(1).fill_(leaky_rate), requires_grad=False))
        # end if
    # end __init__

    ###############################################
    # PUBLIC
    ###############################################

    # Forward
    def forward(self, u, y=None, w_out=None):
        """
        Forward
        :param u: Input signal.
        :return: Resulting hidden states.
        """
        # Time length
        time_length = int(u.size()[1])

        # Number of batches
        n_batches = int(u.size()[0])

        # Outputs
        outputs = Variable(torch.zeros(n_batches, time_length, self.output_dim))
        outputs = outputs.cuda() if self.hidden.is_cuda else outputs

        # For each batch
        for b in range(n_batches):
            # Reset hidden layer
            self.reset_hidden()

            # For each steps
            for t in range(time_length):
                # Current input
                ut = u[b, t]

                # Compute input layer
                u_win = self.w_in.mv(ut)

                # Apply W to x
                x_w = self.w.mv(self.hidden)

                # Feedback or not
                if self.feedbacks and self.training and y is not None:
                    # Current target
                    yt = y[b, t]

                    # Compute feedback layer
                    y_wfdb = self.w_fdb.mv(yt)

                    # Add everything
                    x = u_win + x_w + y_wfdb + self.w_bias
                    # x = u_win + x_w + self.w_bias
                elif self.feedbacks and not self.training and w_out is not None:
                    # Add bias
                    bias_hidden = torch.cat((Variable(torch.ones(1)), self.hidden), dim=0)

                    # Compute past output
                    yt = w_out.t().mv(bias_hidden)

                    # Normalize
                    if self.normalize_feedbacks:
                        yt -= torch.min(yt)
                        yt /= torch.max(yt) - torch.min(yt)
                        yt /= torch.sum(yt)
                    # end if

                    # Compute feedback layer
                    y_wfdb = self.w_fdb.mv(yt)

                    # Add everything
                    x = u_win + x_w + y_wfdb + self.w_bias
                    # x = u_win + x_w + self.w_bias
                else:
                    # Add everything
                    x = u_win + x_w + self.w_bias
                # end if

                # Apply activation function
                x = self.nonlin_func(x)

                # Add to outputs
                self.hidden.data = (self.hidden.mul(1.0 - self.leaky_rate) + x.view(self.output_dim).mul(self.leaky_rate)).data

                # New last state
                outputs[b, t] = self.hidden
            # end for
        # end for

        return outputs
    # end forward

    ###############################################
    # PRIVATE
    ###############################################

# end LiESNCell



# -*- coding: utf-8 -*-
#

# Imports
from .Character import Character
from .Character2Gram import Character2Gram
from .Character3Gram import Character3Gram
from .Compose import Compose
from .Embedding import Embedding
from .FunctionWord import FunctionWord
from .GensimModel import GensimModel
from .GloveVector import GloveVector
from .PartOfSpeech import PartOfSpeech
from .Tag import Tag
from .Token import Token
from .Transformer import Transformer

__all__ = [
    'Character', 'Character2Gram', 'Character3Gram', 'Compose', 'Embedding', 'FunctionWord', 'GensimModel', 'Transformer', 'GloveVector',
    'PartOfSpeech', 'Tag', 'Token'
]



# -*- coding: utf-8 -*-
#
# File : echotorch/nn/HESN.py
# Description : ESN with input pre-trained and used with transfer learning.
# Date : 22 March, 2018
#
# This file is part of EchoTorch.  EchoTorch is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Nils Schaetti <nils.schaetti@unine.ch>

import torch
import torch.sparse
import torch.nn as nn
from .LiESN import LiESN


# ESN with input pre-trained and used with transfer learning
class HESN(object):
    """
    ESN with input pre-trained and used with transfer learning
    """

    # Constructor
    def __init__(self, model, input_dim, hidden_dim, output_dim, spectral_radius=0.9,
                 bias_scaling=0, input_scaling=1.0, w=None, w_in=None, w_bias=None, sparsity=None,
                 input_set=[1.0, -1.0], w_sparsity=None, nonlin_func=torch.tanh, learning_algo='inv', ridge_param=0.0,
                 leaky_rate=1.0, train_leaky_rate=False, feedbacks=False, wfdb_sparsity=None,
                 normalize_feedbacks=False):
        # Embedding layer
        self.mode = model

        # Li-ESN
        self.esn = LiESN(input_dim, hidden_dim, output_dim, spectral_radius, bias_scaling, input_scaling,
                         w, w_in, w_bias, sparsity, input_set, w_sparsity, nonlin_func, learning_algo, ridge_param,
                         leaky_rate, train_leaky_rate, feedbacks, wfdb_sparsity, normalize_feedbacks)
    # end __init__

    ###############################################
    # PROPERTIES
    ###############################################

    # Hidden layer
    @property
    def hidden(self):
        """
        Hidden layer
        :return:
        """
        return self.esn.hidden

    # end hidden

    # Hidden weight matrix
    @property
    def w(self):
        """
        Hidden weight matrix
        :return:
        """
        return self.esn.w

    # end w

    # Input matrix
    @property
    def w_in(self):
        """
        Input matrix
        :return:
        """
        return self.esn.w_in
    # end w_in

    ###############################################
    # PUBLIC
    ###############################################

    # Forward
    def forward(self, u, y=None):
        """
        Forward
        :param x:
        :return:
        """
        # Selected features
        selected_features = self.model(u)

        # ESN
        return self.esn(selected_features, y)
    # end forward

# end HESN



# -*- coding: utf-8 -*-
#
# File : echotorch/nn/ESN.py
# Description : An Echo State Network module.
# Date : 26th of January, 2018
#
# This file is part of EchoTorch.  EchoTorch is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Nils Schaetti <nils.schaetti@unine.ch>

"""
Created on 26 January 2018
@author: Nils Schaetti
"""

# Imports
import torch.sparse
import torch
import torch.nn as nn
from torch.autograd import Variable


# Ridge Regression cell
class RRCell(nn.Module):
    """
    Ridge Regression cell
    """

    # Constructor
    def __init__(self, input_dim, output_dim, ridge_param=0.0, feedbacks=False, with_bias=True, learning_algo='inv'):
        """
        Constructor
        :param input_dim: Inputs dimension.
        :param output_dim: Reservoir size
        """
        super(RRCell, self).__init__()

        # Properties
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.ridge_param = ridge_param
        self.feedbacks = feedbacks
        self.with_bias = with_bias
        self.learning_algo = learning_algo

        # Size
        if self.with_bias:
            self.x_size = input_dim + 1
        else:
            self.x_size = input_dim
        # end if

        # Set it as buffer
        self.register_buffer('xTx', Variable(torch.zeros(self.x_size, self.x_size), requires_grad=False))
        self.register_buffer('xTy', Variable(torch.zeros(self.x_size, output_dim), requires_grad=False))
        self.register_buffer('w_out', Variable(torch.zeros(1, input_dim), requires_grad=False))
    # end __init__

    ###############################################
    # PROPERTIES
    ###############################################

    ###############################################
    # PUBLIC
    ###############################################

    # Reset learning
    def reset(self):
        """
        Reset learning
        :return:
        """
        self.xTx.data = torch.zeros(self.x_size, self.x_size)
        self.xTy.data = torch.zeros(self.x_size, self.output_dim)
        self.w_out.data = torch.zeros(1, self.input_dim)

        # Training mode again
        self.train(True)
    # end reset

    # Output matrix
    def get_w_out(self):
        """
        Output matrix
        :return:
        """
        return self.w_out
    # end get_w_out

    # Forward
    def forward(self, x, y=None):
        """
        Forward
        :param x: Input signal.
        :param y: Target outputs
        :return: Output or hidden states
        """
        # Batch size
        batch_size = x.size()[0]

        # Time length
        time_length = x.size()[1]

        # Add bias
        if self.with_bias:
            x = self._add_constant(x)
        # end if

        # Learning algo
        if self.training:
            for b in range(batch_size):
                self.xTx.data.add_(x[b].t().mm(x[b]).data)
                self.xTy.data.add_(x[b].t().mm(y[b]).data)
            # end for
            return x
        elif not self.training:
            # Outputs
            outputs = Variable(torch.zeros(batch_size, time_length, self.output_dim), requires_grad=False)
            outputs = outputs.cuda() if self.w_out.is_cuda else outputs

            # For each batch
            for b in range(batch_size):
                outputs[b] = torch.mm(x[b], self.w_out)
            # end for

            return outputs
        # end if
    # end forward

    # Finish training
    def finalize(self):
        """
        Finalize training with LU factorization or Pseudo-inverse
        """
        if self.learning_algo == 'inv':
            inv_xTx = self.xTx.inverse()
            # inv_xTx = torch.inverse(self.xTx + self.ridge_param * torch.eye(self._input_dim + self.with_bias))
            self.w_out.data = torch.mm(inv_xTx, self.xTy).data
        else:
            self.w_out.data = torch.gesv(self.xTy, self.xTx + torch.eye(self.esn_cell.output_dim).mul(self.ridge_param)).data
        # end if

        # Not in training mode anymore
        self.train(False)
    # end finalize

    ###############################################
    # PRIVATE
    ###############################################

    # Add constant
    def _add_constant(self, x):
        """
        Add constant
        :param x:
        :return:
        """
        if x.is_cuda:
            bias = Variable(torch.ones((x.size()[0], x.size()[1], 1)).cuda(), requires_grad=False)
        else:
            bias = Variable(torch.ones((x.size()[0], x.size()[1], 1)), requires_grad=False)
        # end if
        return torch.cat((bias, x), dim=2)
    # end _add_constant

# end RRCell



# -*- coding: utf-8 -*-
#

# Imports
import torch
from torch.utils.data.dataset import Dataset
import collections


# Mackey Glass dataset
class MackeyGlassDataset(Dataset):
    """
    Mackey Glass dataset
    """

    # Constructor
    def __init__(self, sample_len, n_samples, tau=17, seed=None):
        """
        Constructor
        :param sample_len: Length of the time-series in time steps.
        :param n_samples: Number of samples to generate.
        :param tau: Delay of the MG with commonly used value of tau=17 (mild chaos) and tau=30 is moderate chaos.
        :param seed: Seed of random number generator.
        """
        # Properties
        self.sample_len = sample_len
        self.n_samples = n_samples
        self.tau = tau
        self.delta_t = 10
        self.timeseries = 1.2
        self.history_len = tau * self.delta_t

        # Init seed if needed
        if seed is not None:
            torch.manual_seed(seed)
        # end if
    # end __init__

    # Length
    def __len__(self):
        """
        Length
        :return:
        """
        return self.n_samples
    # end __len__

    # Get item
    def __getitem__(self, idx):
        """
        Get item
        :param idx:
        :return:
        """
        # History
        history = collections.deque(1.2 * torch.ones(self.history_len) + 0.2 * (torch.rand(self.history_len) - 0.5))

        # Preallocate tensor for time-serie
        inp = torch.zeros(self.sample_len, 1)

        # For each time step
        for timestep in range(self.sample_len):
            for _ in range(self.delta_t):
                xtau = history.popleft()
                history.append(self.timeseries)
                self.timeseries = history[-1] + (0.2 * xtau / (1.0 + xtau ** 10) - 0.1 * history[-1]) / self.delta_t
            # end for
            inp[timestep] = self.timeseries
        # end for

        # Squash timeseries through tanh
        return torch.tan(inp - 1)
    # end __getitem__

# end MackeyGlassDataset



# -*- coding: utf-8 -*-
#
# File : echotorch/nn/ESN.py
# Description : An Echo State Network module.
# Date : 26th of January, 2018
#
# This file is part of EchoTorch.  EchoTorch is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Nils Schaetti, University of Neuchâtel <nils.schaetti@unine.ch>

"""
Created on 26 January 2018
@author: Nils Schaetti
"""

# Imports
import torch.sparse
import torch
import torch.nn as nn
from torch.autograd import Variable
from .BDESNCell import BDESNCell
from RRCell import RRCell


# Bi-directional Echo State Network module
class BDESN(nn.Module):
    """
    Bi-directional Echo State Network module
    """

    # Constructor
    def __init__(self, input_dim, hidden_dim, output_dim, leaky_rate=1.0, spectral_radius=0.9, bias_scaling=0,
                 input_scaling=1.0, w=None, w_in=None, w_bias=None, sparsity=None, input_set=[1.0, -1.0],
                 w_sparsity=None, nonlin_func=torch.tanh, learning_algo='inv', ridge_param=0.0, create_cell=True):
        """
        Constructor
        :param input_dim: Inputs dimension.
        :param hidden_dim: Hidden layer dimension
        :param output_dim: Reservoir size
        :param spectral_radius: Reservoir's spectral radius
        :param bias_scaling: Scaling of the bias, a constant input to each neuron (default: 0, no bias)
        :param input_scaling: Scaling of the input weight matrix, default 1.
        :param w: Internal weights matrix
        :param w_in: Input-reservoir weights matrix
        :param w_bias: Bias weights matrix
        :param sparsity:
        :param input_set:
        :param w_sparsity:
        :param nonlin_func: Reservoir's activation function (tanh, sig, relu)
        :param learning_algo: Which learning algorithm to use (inv, LU, grad)
        """
        super(BDESN, self).__init__()

        # Properties
        self.output_dim = output_dim

        # Recurrent layer
        if create_cell:
            self.esn_cell = BDESNCell(
                input_dim=input_dim, hidden_dim=hidden_dim, spectral_radius=spectral_radius, bias_scaling=bias_scaling,
                input_scaling=input_scaling, w=w, w_in=w_in, w_bias=w_bias, sparsity=sparsity, input_set=input_set,
                w_sparsity=w_sparsity, nonlin_func=nonlin_func, leaky_rate=leaky_rate, create_cell=create_cell
            )
        # end if

        # Ouput layer
        self.output = RRCell(
            input_dim=hidden_dim * 2, output_dim=output_dim, ridge_param=ridge_param, learning_algo=learning_algo
        )
    # end __init__

    ###############################################
    # PROPERTIES
    ###############################################

    # Hidden layer
    @property
    def hidden(self):
        """
        Hidden layer
        :return:
        """
        return self.esn_cell.hidden
    # end hidden

    # Hidden weight matrix
    @property
    def w(self):
        """
        Hidden weight matrix
        :return:
        """
        return self.esn_cell.w
    # end w

    # Input matrix
    @property
    def w_in(self):
        """
        Input matrix
        :return:
        """
        return self.esn_cell.w_in
    # end w_in

    ###############################################
    # PUBLIC
    ###############################################

    # Reset learning
    def reset(self):
        """
        Reset learning
        :return:
        """
        # Reset output layer
        self.output.reset()

        # Training mode again
        self.train(True)
    # end reset

    # Output matrix
    def get_w_out(self):
        """
        Output matrix
        :return:
        """
        return self.output.w_out
    # end get_w_out

    # Set W
    def set_w(self, w):
        """
        Set W
        :param w:
        :return:
        """
        self.esn_cell.w = w
    # end set_w

    # Forward
    def forward(self, u, y=None):
        """
        Forward
        :param u: Input signal.
        :return: Output or hidden states
        """
        # Compute hidden states
        hidden_states = self.esn_cell(u)

        # Learning algorithm
        return self.output(hidden_states, y)
    # end forward

    # Finish training
    def finalize(self):
        """
        Finalize training with LU factorization
        """
        # Finalize output training
        self.output.finalize()

        # Not in training mode anymore
        self.train(False)
    # end finalize

    # Reset hidden layer
    def reset_hidden(self):
        """
        Reset hidden layer
        :return:
        """
        self.esn_cell.reset_hidden()
    # end reset_hidden

    # Get W's spectral radius
    def get_spectral_radius(self):
        """
        Get W's spectral radius
        :return: W's spectral radius
        """
        return self.esn_cell.get_spectral_raduis()
    # end spectral_radius

# end BDESN



# -*- coding: utf-8 -*-
#

# Imports
import torch
import math
from decimal import Decimal
import numpy as np


# Normalized root-mean-square error
def nrmse(outputs, targets):
    """
    Normalized root-mean square error
    :param outputs: Module's outputs
    :param targets: Target signal to be learned
    :return: Normalized root-mean square deviation
    """
    # Flatten tensors
    outputs = outputs.view(outputs.nelement())
    targets = targets.view(targets.nelement())

    # Check dim
    if outputs.size() != targets.size():
        raise ValueError(u"Ouputs and targets tensors don have the same number of elements")
    # end if

    # Normalization with N-1
    var = torch.std(targets) ** 2

    # Error
    error = (targets - outputs) ** 2

    # Return
    return float(math.sqrt(torch.mean(error) / var))
# end nrmse


# Root-mean square error
def rmse(outputs, targets):
    """
    Root-mean square error
    :param outputs: Module's outputs
    :param targets: Target signal to be learned
    :return: Root-mean square deviation
    """
    # Flatten tensors
    outputs = outputs.view(outputs.nelement())
    targets = targets.view(targets.nelement())

    # Check dim
    if outputs.size() != targets.size():
        raise ValueError(u"Ouputs and targets tensors don have the same number of elements")
    # end if

    # Error
    error = (targets - outputs) ** 2

    # Return
    return float(math.sqrt(torch.mean(error)))
# end rmsd


# Mean square error
def mse(outputs, targets):
    """
    Mean square error
    :param outputs: Module's outputs
    :param targets: Target signal to be learned
    :return: Mean square deviation
    """
    # Flatten tensors
    outputs = outputs.view(outputs.nelement())
    targets = targets.view(targets.nelement())

    # Check dim
    if outputs.size() != targets.size():
        raise ValueError(u"Ouputs and targets tensors don have the same number of elements")
    # end if

    # Error
    error = (targets - outputs) ** 2

    # Return
    return float(torch.mean(error))
# end mse


# Normalized mean square error
def nmse(outputs, targets):
    """
    Normalized mean square error
    :param outputs: Module's output
    :param targets: Target signal to be learned
    :return: Normalized mean square deviation
    """
    # Flatten tensors
    outputs = outputs.view(outputs.nelement())
    targets = targets.view(targets.nelement())

    # Check dim
    if outputs.size() != targets.size():
        raise ValueError(u"Ouputs and targets tensors don have the same number of elements")
    # end if

    # Normalization with N-1
    var = torch.std(targets) ** 2

    # Error
    error = (targets - outputs) ** 2

    # Return
    return float(torch.mean(error) / var)
# end nmse


# Perplexity
def perplexity(output_probs, targets, log=False):
    """
    Perplexity
    :param output_probs: Output probabilities for each word/tokens (length x n_tokens)
    :param targets: Real word index
    :return: Perplexity
    """
    pp = Decimal(1.0)
    e_vec = torch.FloatTensor(output_probs.size(0), output_probs.size(1)).fill_(np.e)
    if log:
        set_p = 1.0 / torch.gather(torch.pow(e_vec, exponent=output_probs.data.cpu()), 1,
                                   targets.data.cpu().unsqueeze(1))
    else:
        set_p = 1.0 / torch.gather(output_probs.data.cpu(), 1, targets.data.cpu().unsqueeze(1))
    # end if
    for j in range(set_p.size(0)):
        pp *= Decimal(set_p[j][0])
    # end for
    return pp
# end perplexity


# Cumulative perplexity
def cumperplexity(output_probs, targets, log=False):
    """
    Cumulative perplexity
    :param output_probs:
    :param targets:
    :param log:
    :return:
    """
    # Perplexity
    if log:
        e_vec = torch.FloatTensor(output_probs.size(0), output_probs.size(1)).fill_(np.e)
        set_p = torch.gather(torch.pow(e_vec, exponent=output_probs.data.cpu()), 1,
                             targets.data.cpu().unsqueeze(1))
    else:
        set_p = torch.gather(output_probs.data.cpu(), 1, targets.data.cpu().unsqueeze(1))
    # end if
    return np.sum(np.log2(set_p.numpy()))
# end cumperplexity



# -*- coding: utf-8 -*-
#

# Imports
import torch
import spacy
from .Transformer import Transformer


# Transform text to part-of-speech vectors
class PartOfSpeech(Transformer):
    """
    Transform text to part-of-speech vectors
    """

    # Constructor
    def __init__(self, model="en_core_web_lg"):
        """
        Constructor
        :param model: Spacy's model to load.
        """
        # Super constructor
        super(PartOfSpeech, self).__init__()

        # Properties
        self.model = model
        self.nlp = spacy.load(model)
    # end __init__

    ##############################################
    # Public
    ##############################################

    # Get tags
    def get_tags(self):
        """
        Get tags.
        :return: A list of tags.
        """
        return [u"ADJ", u"ADP", u"ADV", u"CCONJ", u"DET", u"INTJ", u"NOUN", u"NUM", u"PART", u"PRON", u"PROPN",
                u"PUNCT", u"SYM", u"VERB", u"SPACE", u"X"]
    # end get_tags

    ##############################################
    # Override
    ##############################################

    # Convert a string
    def __call__(self, text):
        """
        Convert a string to a ESN input
        :param text: Text to convert
        :return: Tensor of word vectors
        """
        # Inputs as tensor
        inputs = torch.FloatTensor(1, self.input_dim)

        # Start
        start = True

        # For each tokens
        for token in self.nlp(text):
            pos = self.tag_to_symbol(token.pos_)

            if not start:
                inputs = torch.cat((inputs, pos), dim=0)
            else:
                inputs = pos
                start = False
            # end if
        # end for

        return inputs, inputs.size()[0]
    # end convert

# end PartOfSpeech



# -*- coding: utf-8 -*-
#

# Imports
import torch
import spacy
from .Transformer import Transformer


# Transform text to tag vectors
class Tag(Transformer):
    """
    Transform text to tag vectors
    """

    # Constructor
    def __init__(self, model="en_core_web_lg"):
        """
        Constructor
        :param model: Spacy's model to load.
        """
        # Super constructor
        super(Tag, self).__init__()

        # Properties
        self.model = model
        self.nlp = spacy.load(model)
    # end __init__

    ##############################################
    # Public
    ##############################################

    # Get tags
    def get_tags(self):
        """
        Get all tags.
        :return: A list of tags.
        """
        return [u"''", u",", u":", u".", u"``", u"-LRB-", u"-RRB-", u"AFX", u"CC", u"CD", u"DT", u"EX", u"FW",
                u"IN", u"JJ", u"JJR", u"JJS", u"LS", u"MD", u"NN", u"NNS", u"NNP", u"NNPS", u"PDT", u"POS", u"PRP",
                u"PRP$", u"RB", u"RBR", u"RBS", u"RP", u"SYM", u"TO", u"UH", u"VB", u"VBZ", u"VBP", u"VBD", u"VBN",
                u"VBG", u"WDT", u"WP", u"WP$", u"WRB", u"X"]
    # end get_tags

    ##############################################
    # Override
    ##############################################

    # Convert a string
    def __call__(self, text):
        """
        Convert a string to a ESN input
        :param text: Text to convert
        :return: Tensor of word vectors
        """
        # Inputs as tensor
        inputs = torch.FloatTensor(1, self.input_dim)

        # Null symbol
        null_symbol = torch.zeros(1, self.input_dim)
        null_symbol[0, -1] = 1.0

        # Start
        start = True

        # For each tokens
        for token in self.nlp(text):
            # Replace if not function word
            if token.tag_ not in self.symbols:
                token_tag = u"X"
            else:
                token_tag = token.tag_
            # end if

            # Get tag
            tag = self.tag_to_symbol(token_tag)

            # Add
            if not start:
                inputs = torch.cat((inputs, tag), dim=0)
            else:
                inputs = tag
                start = False
            # end if
        # end for

        return inputs, inputs.size()[0]
    # end convert

# end FunctionWord



# -*- coding: utf-8 -*-
#

# Imports
from .error_measures import nrmse, nmse, rmse, mse, perplexity, cumperplexity
from .utility_functions import spectral_radius, normalize, average_prob, max_average_through_time

__all__ = [
    'nrmse', 'nmse', 'rmse', 'mse', 'perplexity', 'cumperplexity', 'spectral_radius', 'normalize', 'average_prob', 'max_average_through_time'
]



# -*- coding: utf-8 -*-
#

# Imports
import spacy


# Transform text to a list of tokens
class Token(object):
    """
    Transform text to a list of tokens
    """

    # Constructor
    def __init__(self, model="en_core_web_lg"):
        """
        Constructor
        :param model: Spacy's model to load.
        """
        # Properties
        self.model = model
        self.nlp = spacy.load(model)
    # end __init__

    ##############################################
    # Properties
    ##############################################

    # Get the number of inputs
    @property
    def input_dim(self):
        """
        Get the number of inputs.
        :return: The input size.
        """
        return 1
    # end input_dim

    ##############################################
    # Override
    ##############################################

    # Convert a string
    def __call__(self, text):
        """
        Convert a string to a ESN input
        :param text: Text to convert
        :return: Tensor of word vectors
        """
        # Inputs as a list
        tokens = list()

        # For each tokens
        for token in self.nlp(text):
            tokens.append(unicode(token.text))
        # end for

        return tokens, len(tokens)
    # end convert

    ##############################################
    # Private
    ##############################################

    # Get inputs size
    def _get_inputs_size(self):
        """
        Get inputs size.
        :return:
        """
        return 1
    # end if

    ##############################################
    # Static
    ##############################################

# end Token



# -*- coding: utf-8 -*-
#

# Imports
import torch
from torch.utils.data.dataset import Dataset
import numpy as np


# Switch attractor dataset
class SwitchAttractorDataset(Dataset):
    """
    Generate a dataset where the reservoir must switch
    between two attractors.
    """

    # Constructor
    def __init__(self, sample_len, n_samples, seed=None):
        """
        Constructor
        :param sample_len: Length of the time-series in time steps.
        :param n_samples: Number of samples to generate.
        :param system_order: th order NARMA
        :param seed: Seed of random number generator.
        """
        # Properties
        self.sample_len = sample_len
        self.n_samples = n_samples

        # Init seed if needed
        if seed is not None:
            torch.manual_seed(seed)
        # end if

        # Generate data set
        self.inputs, self.outputs = self._generate()
    # end __init__

    #############################################
    # OVERRIDE
    #############################################

    # Length
    def __len__(self):
        """
        Length
        :return:
        """
        return self.n_samples
    # end __len__

    # Get item
    def __getitem__(self, idx):
        """
        Get item
        :param idx:
        :return:
        """
        return self.inputs[idx], self.outputs[idx]
    # end __getitem__

    ##############################################
    # PRIVATE
    ##############################################

    # Generate
    def _generate(self):
        """
        Generate dataset
        :return:
        """
        inputs = list()
        outputs = list()

        # Generate each sample
        for i in range(self.n_samples):
            # Start end stop
            start = np.random.randint(0, self.sample_len)
            stop = np.random.randint(start, start + self.sample_len / 2)

            # Limits
            if stop >= self.sample_len:
                stop = self.sample_len - 1
            # end if

            # Sample tensor
            inp = torch.zeros(self.sample_len, 1)
            out = torch.zeros(self.sample_len)

            # Set inputs
            inp[start, 0] = 1.0
            inp[stop] = 1.0

            # Set outputs
            out[start:stop] = 1.0

            # Add
            inputs.append(inp)
            outputs.append(out)
        # end for

        return inputs, outputs
    # end _generate

# end SwitchAttractorDataset



# -*- coding: utf-8 -*-
#
# File : echotorch/nn/ESN.py
# Description : An Echo State Network module.
# Date : 26th of January, 2018
#
# This file is part of EchoTorch.  EchoTorch is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Nils Schaetti <nils.schaetti@unine.ch>

"""
Created on 26 January 2018
@author: Nils Schaetti
"""

import torch
from .LiESNCell import LiESNCell
from .ESN import ESN


# Leaky-Integrated Echo State Network module
class LiESN(ESN):
    """
    Leaky-Integrated Echo State Network module
    """

    # Constructor
    def __init__(self, input_dim, hidden_dim, output_dim, spectral_radius=0.9,
                 bias_scaling=0, input_scaling=1.0, w=None, w_in=None, w_bias=None, sparsity=None,
                 input_set=[1.0, -1.0], w_sparsity=None, nonlin_func=torch.tanh, learning_algo='inv', ridge_param=0.0,
                 leaky_rate=1.0, train_leaky_rate=False, feedbacks=False, wfdb_sparsity=None,
                 normalize_feedbacks=False):
        """
        Constructor
        :param input_dim:
        :param hidden_dim:
        :param output_dim:
        :param spectral_radius:
        :param bias_scaling:
        :param input_scaling:
        :param w:
        :param w_in:
        :param w_bias:
        :param sparsity:
        :param input_set:
        :param w_sparsity:
        :param nonlin_func:
        :param learning_algo:
        :param ridge_param:
        :param leaky_rate:
        :param train_leaky_rate:
        :param feedbacks:
        """
        super(LiESN, self).__init__(input_dim, hidden_dim, output_dim, spectral_radius=spectral_radius,
                                    bias_scaling=bias_scaling, input_scaling=input_scaling,
                                    w=w, w_in=w_in, w_bias=w_bias, sparsity=sparsity, input_set=input_set,
                                    w_sparsity=w_sparsity, nonlin_func=nonlin_func, learning_algo=learning_algo,
                                    ridge_param=ridge_param, create_cell=False, feedbacks=feedbacks,
                                    wfdb_sparsity=wfdb_sparsity, normalize_feedbacks=normalize_feedbacks)

        # Recurrent layer
        self.esn_cell = LiESNCell(leaky_rate, train_leaky_rate, input_dim, hidden_dim, spectral_radius=spectral_radius,
                                  bias_scaling=bias_scaling, input_scaling=input_scaling,
                                  w=w, w_in=w_in, w_bias=w_bias, sparsity=sparsity, input_set=input_set,
                                  w_sparsity=w_sparsity, nonlin_func=nonlin_func, feedbacks=feedbacks,
                                  feedbacks_dim=output_dim, wfdb_sparsity=wfdb_sparsity,
                                  normalize_feedbacks=normalize_feedbacks)
    # end __init__

    ###############################################
    # PROPERTIES
    ###############################################

    ###############################################
    # PUBLIC
    ###############################################

    ###############################################
    # PRIVATE
    ###############################################

# end ESNCell



# -*- coding: utf-8 -*-
#

# Imports
import torch
from torch.utils.data.dataset import Dataset


# Generates a series of input timeseries and delayed versions as outputs.
class MemTestDataset(Dataset):
    """
    Generates a series of input timeseries and delayed versions as outputs.
    Delay is given in number of timesteps. Can be used to empirically measure the
    memory capacity of a system.
    """

    # Constructor
    def __init__(self, sample_len, n_samples, n_delays=10, seed=None):
        """
        Constructor
        :param sample_len: Length of the time-series in time steps.
        :param n_samples: Number of samples to generate.
        :param n_delays: Number of step to delay
        :param seed: Seed of random number generator.
        """
        # Properties
        self.sample_len = sample_len
        self.n_samples = n_samples
        self.n_delays = n_delays

        # Init seed if needed
        if seed is not None:
            torch.manual_seed(seed)
        # end if
    # end __init__

    # Length
    def __len__(self):
        """
        Length
        :return:
        """
        return self.n_samples
    # end __len__

    # Get item
    def __getitem__(self, idx):
        """
        Get item
        :param idx:
        :return:
        """
        inputs = (torch.rand(self.sample_len, 1) - 0.5) * 1.6
        outputs = torch.zeros(self.sample_len, self.n_delays)
        for k in range(self.n_delays):
            outputs[:, k:k+1] = torch.cat((torch.zeros(k + 1, 1), inputs[:-k - 1, :]), dim=0)
        # end for
        return inputs, outputs
    # end __getitem__

# end MemTestDataset



# -*- coding: utf-8 -*-
#
# File : echotorch/models/NilsNet.py
# Description : An NilsNet module.
# Date : 09th of April, 2018
#
# This file is part of EchoTorch.  EchoTorch is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Nils Schaetti, University of Neuchâtel <nils.schaetti@unine.ch>

# Imports
import torchvision
import torch.nn as nn
import torch.legacy.nn as lnn
from echotorch import nn as ecnn


# A NilsNet
class NilsNet(nn.Module):
    """
    A NilsNet
    """

    # Constructor
    def __init__(self, reservoir_dim, sfa_dim, ica_dim, pretrained=False, feature_selector='resnet18'):
        """
        Constructor
        """
        # Upper class
        super(NilsNet, self).__init__()

        # ResNet
        if feature_selector == 'resnet18':
            self.feature_selector = torchvision.models.resnet18(pretrained=True)
        elif feature_selector == 'resnet34':
            self.feature_selector = torchvision.models.resnet34(pretrained=True)
        elif feature_selector == 'resnet50':
            self.feature_selector = torchvision.models.resnet50(pretrained=True)
        elif feature_selector == 'alexnet':
            self.feature_selector = torchvision.models.alexnet(pretrained=True)
        # end if

        # Skip last layer
        self.reservoir_input_dim = self.feature_selector.fc.in_features
        self.feature_selector.fc = ecnn.Identity()

        # Echo State Network
        # self.esn = ecnn.ESNCell(input_dim=self.reservoir_input_dim, output_dim=reservoir_dim)

        # Slow feature analysis layer
        # self.sfa = ecnn.SFACell(input_dim=reservoir_dim, output_dim=sfa_dim)

        # Independent Feature Analysis layer
        # self.ica = ecnn.ICACell(input_dim=sfa_dim, output_dim=ica_dim)
    # end __init__

    # Forward
    def forward(self, x):
        """
        Forward
        :return:
        """
        # ResNet
        return self.feature_selector(x)
    # end forward

# end NilsNet



# -*- coding: utf-8 -*-
#

# Imports
import gensim
from gensim.utils import tokenize
import torch
import numpy as np


# Transform text to vectors with a Gensim model
class GensimModel(object):
    """
    Transform text to vectors with a Gensim model
    """

    # Constructor
    def __init__(self, model_path):
        """
        Constructor
        :param model_path: Model's path.
        """
        # Properties
        self.model_path = model_path

        # Format
        binary = False if model_path[-4:] == ".vec" else True

        # Load
        self.model = gensim.models.KeyedVectors.load_word2vec_format(model_path, binary=binary, unicode_errors='ignore')

        # OOV
        self.oov = 0.0
    # end __init__

    ##############################################
    # Properties
    ##############################################

    # Get the number of inputs
    @property
    def input_dim(self):
        """
        Get the number of inputs.
        :return: The input size.
        """
        return 300
    # end input_dim

    ##############################################
    # Override
    ##############################################

    # Convert a string
    def __call__(self, text):
        """
        Convert a string to a ESN input
        :param text: Text to convert
        :return: Tensor of word vectors
        """
        # Inputs as tensor
        inputs = torch.FloatTensor(1, self.input_dim)

        # Start
        start = True
        count = 0.0

        # OOV
        zero = 0.0
        self.oov = 0.0

        # For each tokens
        for token in tokenize(text):
            found = False
            # Try normal
            try:
                word_vector = self.model[token]
                found = True
            except KeyError:
                pass
            # end try

            # Try lower
            if not found:
                try:
                    word_vector = self.model[token.lower()]
                except KeyError:
                    zero += 1.0
                    word_vector = np.zeros(self.input_dim)
                # end try
            # end if

            # Start/continue
            if not start:
                inputs = torch.cat((inputs, torch.FloatTensor(word_vector).unsqueeze_(0)), dim=0)
            else:
                inputs = torch.FloatTensor(word_vector).unsqueeze_(0)
                start = False
            # end if
            count += 1
        # end for

        # OOV
        self.oov = zero / count * 100.0

        return inputs, inputs.size()[0]
    # end convert

    ##############################################
    # Static
    #########################################


# -*- coding: utf-8 -*-
#

# Imports
import torch
from torch.utils.data.dataset import Dataset
import numpy as np


# Logistic Map dataset
class LogisticMapDataset(Dataset):
    """
    Logistic Map dataset
    """

    # Constructor
    def __init__(self, sample_len, n_samples, alpha=5, beta=11, gamma=13, c=3.6, b=0.13, seed=None):
        """
        Constructor
        :param sample_len:
        :param n_samples:
        :param alpha:
        :param beta:
        :param gamma:
        :param c:
        :param b:
        :param seed:
        """
        # Properties
        self.sample_len = sample_len
        self.n_samples = n_samples
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma
        self.c = c
        self.b = b
        self.p2 = np.pi * 2

        # Init seed if needed
        if seed is not None:
            torch.manual_seed(seed)
        # end if
    # end __init__

    # Length
    def __len__(self):
        """
        Length
        :return:
        """
        return self.n_samples
    # end __len__

    # Get item
    def __getitem__(self, idx):
        """
        Get item
        :param idx:
        :return:
        """
        # Time and forces
        t = np.linspace(0, 1, self.sample_len, endpoint=0)
        dforce = np.sin(self.p2 * self.alpha * t) + np.sin(self.p2 * self.beta * t) + np.sin(self.p2 * self.gamma * t)

        # Series
        series = torch.zeros(self.sample_len, 1)
        series[0] = 0.6

        # Generate
        for i in range(1, self.sample_len):
            series[i] = self._logistic_map(series[i-1], self.c + self.b * dforce[i])
        # end for

        return series
    # end __getitem__

    #######################################
    # Private
    #######################################

    # Logistic map
    def _logistic_map(self, x, r):
        """
        Logistic map
        :param x:
        :param r:
        :return:
        """
        return r * x * (1-x)
    # end logistic_map

# end MackeyGlassDataset



# -*- coding: utf-8 -*-
#

# Imports
import torch
import spacy
from .Transformer import Transformer


# Transform text to a function word vectors
class FunctionWord(Transformer):
    """
    Transform text to character vectors
    """

    # Constructor
    def __init__(self, model="en_core_web_lg"):
        """
        Constructor
        :param model: Spacy's model to load.
        """
        # Super constructor
        super(FunctionWord, self).__init__()

        # Properties
        self.model = model
        self.nlp = spacy.load(model)
    # end __init__

    ##############################################
    # Public
    ##############################################

    # Get tags
    def get_tags(self):
        """
        Get tags.
        :return: A tag list.
        """
        return [u"a", u"about", u"above", u"after", u"after", u"again", u"against", u"ago", u"ahead",
                u"all",
                u"almost", u"along", u"already", u"also", u"although", u"always", u"am", u"among", u"an",
                u"and", u"any", u"are", u"aren't", u"around", u"as", u"at", u"away", u"backward",
                u"backwards", u"be", u"because", u"before", u"behind", u"below", u"beneath", u"beside",
                u"between", u"both", u"but", u"by", u"can", u"cannot", u"can't", u"cause", u"'cos",
                u"could",
                u"couldn't", u"'d", u"despite", u"did", u"didn't", u"do", u"does", u"doesn't", u"don't",
                u"down", u"during", u"each", u"either", u"even", u"ever", u"every", u"except", u"for",
                u"forward", u"from", u"had", u"hadn't", u"has", u"hasn't", u"have", u"haven't", u"he",
                u"her", u"here", u"hers", u"herself", u"him", u"himself", u"his", u"how", u"however",
                u"I",
                u"if", u"in", u"inside", u"inspite", u"instead", u"into", u"is", u"isn't", u"it", u"its",
                u"itself", u"just", u"'ll", u"least", u"less", u"like", u"'m", u"many", u"may",
                u"mayn't",
                u"me", u"might", u"mightn't", u"mine", u"more", u"most", u"much", u"must", u"mustn't",
                u"my", u"myself", u"near", u"need", u"needn't", u"needs", u"neither", u"never", u"no",
                u"none", u"nor", u"not", u"now", u"of", u"off", u"often", u"on", u"once", u"only",
                u"onto",
                u"or", u"ought", u"oughtn't", u"our", u"ours", u"ourselves", u"out", u"outside", u"over",
                u"past", u"perhaps", u"quite", u"'re", u"rather", u"'s", u"seldom", u"several", u"shall",
                u"shan't", u"she", u"should", u"shouldn't", u"since", u"so", u"some", u"sometimes",
                u"soon",
                u"than", u"that", u"the", u"their", u"theirs", u"them", u"themselves", u"then", u"there",
                u"therefore", u"these", u"they", u"this", u"those", u"though", u"through", u"thus",
                u"till",
                u"to", u"together", u"too", u"towards", u"under", u"unless", u"until", u"up", u"upon",
                u"us", u"used", u"usedn't", u"usen't", u"usually", u"'ve", u"very", u"was", u"wasn't",
                u"we", u"well", u"were", u"weren't", u"what", u"when", u"where", u"whether", u"which",
                u"while", u"who", u"whom", u"whose", u"why", u"will", u"with", u"without", u"won't",
                u"would", u"wouldn't", u"yet", u"you", u"your", u"yours", u"yourself", u"yourselves", u"X"]
    # end get_tags

    ##############################################
    # Override
    ##############################################

    # Convert a string
    def __call__(self, text):
        """
        Convert a string to a ESN input
        :param text: Text to convert
        :return: Tensor of word vectors
        """
        # Inputs as tensor
        inputs = torch.FloatTensor(1, self.input_dim)

        # Null symbol
        null_symbol = torch.zeros(1, self.input_dim)
        null_symbol[0, -1] = 1.0

        # Start
        start = True

        # For each tokens
        for token in self.nlp(text):
            # Replace if not function word
            if token.text not in self.symbols:
                token_fw = u"X"
            else:
                token_fw = token.text
            # end if

            # Get tag
            fw = self.tag_to_symbol(token_fw)

            # Add
            if not start:
                inputs = torch.cat((inputs, fw), dim=0)
            else:
                inputs = fw
                start = False
            # end if
        # end for

        return inputs, inputs.size()[0]
    # end convert

# end FunctionWord



# -*- coding: utf-8 -*-
#

# Imports
import torch


# Compute spectral radius of a square 2-D tensor
def spectral_radius(m):
    """
    Compute spectral radius of a square 2-D tensor
    :param m: squared 2D tensor
    :return:
    """
    return torch.max(torch.abs(torch.eig(m)[0]))
# end nrmse


# Normalize a tensor on a single dimension
def normalize(tensor, dim=1):
    """
    Normalize a tensor on a single dimension
    :param t:
    :return:
    """
    pass
# end normalize


# Average probabilties through time
def average_prob(tensor, dim=0):
    """
    Average probabilities through time
    :param tensor:
    :param dim:
    :return:
    """
    return torch.mean(tensor, dim=dim)
# end average_prob


# Max average through time
def max_average_through_time(tensor, dim=0):
    """
    Max average through time
    :param tensor:
    :param dim: Time dimension
    :return:
    """
    average = torch.mean(tensor, dim=dim)
    return torch.max(average, dim=dim)[1]
# end max_average_through_time



# -*- coding: utf-8 -*-
#

# Imports
import gensim
from gensim.utils import tokenize
import torch
import numpy as np


# Transform text to vectors with embedding
class Embedding(object):
    """
    Transform text to vectors with embedding
    """

    # Constructor
    def __init__(self, weights):
        """
        Constructor
        :param weights: Embedding weight matrix
        """
        # Properties
        self.weights = weights
        self.voc_size = weights.size(0)
        self.embedding_dim = weights.size(1)
    # end __init__

    ##############################################
    # Properties
    ##############################################

    # Get the number of inputs
    @property
    def input_dim(self):
        """
        Get the number of inputs
        :return:
        """
        return self.embedding_dim
    # end input_dim

    ##############################################
    # Override
    ##############################################

    # Convert a string
    def __call__(self, idxs):
        """
        Convert a strng
        :param text:
        :return:
        """
        # Inputs as tensor
        inputs = torch.FloatTensor(1, self.embedding_dim)

        # Start
        start = True
        count = 0.0

        # OOV
        zero = 0.0
        self.oov = 0.0

        # For each inputs
        for i in range(idxs.size(0)):
            # Get token ix
            ix = idxs[i]

            # Get vector
            if ix < self.voc_size:
                embedding_vector = self.weights[ix]
            else:
                embedding_vector = torch.zeros(self.embedding_dim)
            # end if

            # Test zero
            if torch.sum(embedding_vector) == 0.0:
                zero += 1.0
                embedding_vector = np.zeros(self.input_dim)
            # end if

            # Start/continue
            if not start:
                inputs = torch.cat((inputs, torch.FloatTensor(embedding_vector).unsqueeze_(0)), dim=0)
            else:
                inputs = torch.FloatTensor(embedding_vector).unsqueeze_(0)
                start = False
            # end if
            count += 1
        # end for

        # OOV
        self.oov = zero / count * 100.0

        return inputs, inputs.size()[0]
    # end convert

    ##############################################
    # Static
    ##############################################


# end Embedding



# -*- coding: utf-8 -*-
#

# Imports
import torch
import spacy
import numpy as np
from datetime import datetime


# Transform text to word vectors
class GloveVector(object):
    """
    Transform text to word vectors
    """

    # Constructor
    def __init__(self, model="en_vectors_web_lg"):
        """
        Constructor
        :param model: Spacy's model to load.
        """
        # Properties
        self.model = model
        self.nlp = spacy.load(model)
        self.oov = 0.0
    # end __init__

    ##############################################
    # Properties
    ##############################################

    # Get the number of inputs
    @property
    def input_dim(self):
        """
        Get the number of inputs.
        :return: The input size.
        """
        return 300
    # end input_dim

    ##############################################
    # Override
    ##############################################

    # Convert a string
    def __call__(self, text):
        """
        Convert a string to a ESN input
        :param text: Text to convert
        :return: Tensor of word vectors
        """
        # Inputs as tensor
        inputs = torch.FloatTensor(1, self.input_dim)

        # Start
        start = True
        count = 0.0

        # Zero count
        zero = 0.0
        self.oov = 0.0

        # For each tokens
        for token in self.nlp(text):
            if np.sum(token.vector) == 0:
                zero += 1.0
            # end if
            if not start:
                inputs = torch.cat((inputs, torch.FloatTensor(token.vector).unsqueeze_(0)), dim=0)
            else:
                inputs = torch.FloatTensor(token.vector).unsqueeze_(0)
                start = False
            # end if
            count += 1.0
        # end for

        # OOV
        self.oov = zero / count * 100.0

        return inputs, inputs.size()[0]
    # end convert

    ##############################################
    # Static
    ##############################################

# end GloveVector



# -*- coding: utf-8 -*-
#

# Imports
from .BDESN import BDESN
from .BDESNPCA import BDESNPCA
from .BDESNCell import BDESNCell
from .ESNCell import ESNCell
from .ESN import ESN
from .LiESNCell import LiESNCell
from .LiESN import LiESN
from .GatedESN import GatedESN
from .ICACell import ICACell
from .Identity import Identity
from .PCACell import PCACell
from .RRCell import RRCell
from .SFACell import SFACell
from .StackedESN import StackedESN

__all__ = [
    'BDESN', 'BDESNPCA', 'BDESNCell', 'ESNCell', 'ESN', 'LiESNCell', 'LiESN', 'GatedESN', 'ICACell', 'Identity',
    'PCACell', 'RRCell', 'SFACell', 'StackedESN'
]



# -*- coding: utf-8 -*-
#

# Imports
import torch


# Base class for text transformers
class Transformer(object):
    """
    Base class for text transformers
    """

    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Properties
        self.symbols = self.generate_symbols()
    # end __init__

    ##############################################
    # Properties
    ##############################################

    # Get the number of inputs
    @property
    def input_dim(self):
        """
        Get the number of inputs.
        :return: The input size.
        """
        return len(self.get_tags())
    # end input_dim

    ##############################################
    # Public
    ##############################################

    # Get tags
    def get_tags(self):
        """
        Get tags.
        :return: A list of tags.
        """
        return []
    # end get_tags

    # Get symbol from tag
    def tag_to_symbol(self, tag):
        """
        Get symbol from tag.
        :param tag: Tag.
        :return: The corresponding symbols.
        """
        if tag in self.symbols.keys():
            return self.symbols[tag]
        return None
    # end word_to_symbol

    # Generate symbols
    def generate_symbols(self):
        """
        Generate word symbols.
        :return: Dictionary of tag to symbols.
        """
        result = dict()
        for index, p in enumerate(self.get_tags()):
            result[p] = torch.zeros(1, self.input_dim)
            result[p][0, index] = 1.0
        # end for
        return result
    # end generate_symbols

    ##############################################
    # Override
    ##############################################

    # Convert a string
    def __call__(self, tokens):
        """
        Convert a string to a ESN input
        :param tokens: Text to convert
        :return: A list of symbols
        """
        pass
    # end convert

    ##############################################
    # Static
    ##############################################

# end TextTransformer



from setuptools import setup, find_packages

setup(name='EchoTorch',
      version='0.1.1',
      description="A Python toolkit for Reservoir Computing.",
      long_description="A Python toolkit for Reservoir Computing and Echo State Network experimentation based on pyTorch.",
      author='Nils Schaetti',
      author_email='nils.schaetti@unine.ch',
      license='GPLv3',
      packages=find_packages(),
      install_requires=[
          'torch',
          'numpy',
          'torchvision'
      ],
      zip_safe=False
      )




# -*- coding: utf-8 -*-
#

# Imports
import torch
from .Transformer import Transformer
import numpy as np


# Transform text to character 2-gram
class Character2Gram(Transformer):
    """
    Transform text to character 2-grams
    """

    # Constructor
    def __init__(self, uppercase=False, gram_to_ix=None, start_ix=0, fixed_length=-1, overlapse=True):
        """
        Constructor
        """
        # Gram to ix
        if gram_to_ix is not None:
            self.gram_count = len(gram_to_ix.keys())
            self.gram_to_ix = gram_to_ix
        else:
            self.gram_count = start_ix
            self.gram_to_ix = dict()
        # end if

        # Ix to gram
        self.ix_to_gram = dict()
        if gram_to_ix is not None:
            for gram in gram_to_ix.keys():
                self.ix_to_gram[gram_to_ix[gram]] = gram
            # end for
        # end if

        # Properties
        self.uppercase = uppercase
        self.fixed_length = fixed_length
        self.overlapse = overlapse

        # Super constructor
        super(Character2Gram, self).__init__()
    # end __init__

    ##############################################
    # Public
    ##############################################

    ##############################################
    # Properties
    ##############################################

    # Get the number of inputs
    @property
    def input_dim(self):
        """
        Get the number of inputs.
        :return: The input size.
        """
        return 1
    # end input_dim

    # Vocabulary size
    @property
    def voc_size(self):
        """
        Vocabulary size
        :return:
        """
        return self.gram_count
    # end voc_size

    ##############################################
    # Private
    ##############################################

    # To upper
    def to_upper(self, gram):
        """
        To upper
        :param gram:
        :return:
        """
        if not self.uppercase:
            return gram.lower()
        # end if
        return gram
    # end to_upper

    ##############################################
    # Override
    ##############################################

    # Convert a string
    def __call__(self, text):
        """
        Convert a string to a ESN input
        :param text: Text to convert
        :return: Tensor of word vectors
        """
        # Step
        if self.overlapse:
            step = 1
        else:
            step = 2
        #  end if

        # Add to voc
        for i in np.arange(0, len(text) - 1, step):
            gram = self.to_upper(text[i] + text[i+1])
            if gram not in self.gram_to_ix.keys():
                self.gram_to_ix[gram] = self.gram_count
                self.ix_to_gram[self.gram_count] = gram
                self.gram_count += 1
            # end if
        # end for

        # List of character to 2grams
        text_idxs = [self.gram_to_ix[self.to_upper(text[i] + text[i+1])] for i in range(len(text)-1)]

        # To long tensor
        text_idxs = torch.LongTensor(text_idxs)

        # Check length
        if self.fixed_length != -1:
            if text_idxs.size(0) > self.fixed_length:
                text_idxs = text_idxs[:self.fixed_length]
            elif text_idxs.size(0) < self.fixed_length:
                zero_idxs = torch.LongTensor(self.fixed_length).fill_(0)
                zero_idxs[:text_idxs.size(0)] = text_idxs
                text_idxs = zero_idxs
            # end if
        # end if

        return text_idxs, text_idxs.size(0)
    # end convert

# end Character2Gram



# -*- coding: utf-8 -*-
#

# Imports
from .Transformer import Transformer


# Compose multiple transformations
class Compose(Transformer):
    """
    Compose multiple transformations
    """

    # Constructor
    def __init__(self, transforms):
        """
        Constructor
        """
        # Properties
        self.transforms = transforms

        # Super constructor
        super(Compose, self).__init__()
    # end __init__

    ##############################################
    # Public
    ##############################################

    ##############################################
    # Properties
    ##############################################

    # Get the number of inputs
    @property
    def input_dim(self):
        """
        Get the number of inputs.
        :return: The input size.
        """
        return self.transforms[-1].input_dim
    # end input_dim

    ##############################################
    # Override
    ##############################################

    # Convert a string
    def __call__(self, text):
        """
        Convert a string to a ESN input
        :param text: Text to convert
        :return: Tensor of word vectors
        """
        # For each transform
        for index, transform in enumerate(self.transforms):
            # Transform
            if index == 0:
                outputs, size = transform(text)
            else:
                outputs, size = transform(outputs)
            # end if
        # end for

        return outputs, size
    # end convert

# end Compose



# -*- coding: utf-8 -*-
#
# File : echotorch/nn/ESN.py
# Description : An Echo State Network module.
# Date : 26th of January, 2018
#
# This file is part of EchoTorch.  EchoTorch is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Nils Schaetti, University of Neuchâtel <nils.schaetti@unine.ch>

"""
Created on 26 January 2018
@author: Nils Schaetti
"""

# Imports
import torch.sparse
import torch
import torch.nn as nn
from torch.autograd import Variable
from . import LiESNCell
from RRCell import RRCell
from ESNCell import ESNCell


# Stacked Echo State Network module
class StackedESN(nn.Module):
    """
    Stacked Echo State Network module
    """

    # Constructor
    def __init__(self, input_dim, hidden_dim, output_dim, leaky_rate=1.0, spectral_radius=0.9, bias_scaling=0,
                 input_scaling=1.0, w=None, w_in=None, w_bias=None, sparsity=None, input_set=(1.0, -1.0),
                 w_sparsity=None, nonlin_func=torch.tanh, learning_algo='inv', ridge_param=0.0, with_bias=True):
        """
        Constructor
        :param input_dim: Inputs dimension.
        :param hidden_dim: Hidden layer dimension
        :param output_dim: Reservoir size
        :param spectral_radius: Reservoir's spectral radius
        :param bias_scaling: Scaling of the bias, a constant input to each neuron (default: 0, no bias)
        :param input_scaling: Scaling of the input weight matrix, default 1.
        :param w: Internation weights matrix
        :param w_in: Input-reservoir weights matrix
        :param w_bias: Bias weights matrix
        :param w_fdb: Feedback weights matrix
        :param sparsity:
        :param input_set:
        :param w_sparsity:
        :param nonlin_func: Reservoir's activation function (tanh, sig, relu)
        :param learning_algo: Which learning algorithm to use (inv, LU, grad)
        """
        super(StackedESN, self).__init__()

        # Properties
        self.n_layers = len(hidden_dim)
        self.esn_layers = list()

        # Number of features
        self.n_features = 0

        # Recurrent layer
        for n in range(self.n_layers):
            # Input dim
            layer_input_dim = input_dim if n == 0 else hidden_dim[n-1]
            self.n_features += hidden_dim[n]

            # Parameters
            layer_leaky_rate = leaky_rate[n] if type(leaky_rate) is list else leaky_rate
            layer_spectral_radius = spectral_radius[n] if type(spectral_radius) is list else spectral_radius
            layer_bias_scaling = bias_scaling[n] if type(bias_scaling) is list else bias_scaling
            layer_input_scaling = input_scaling[n] if type(input_scaling) is list else input_scaling

            # W
            if type(w) is torch.Tensor and w.ndim == 3:
                layer_w = w[n]
            elif type(w) is torch.Tensor:
                layer_w = w
            else:
                layer_w = None
            # end if

            # W in
            if type(w_in) is torch.Tensor and w_in.ndim == 3:
                layer_w_in = w_in[n]
            elif type(w_in) is torch.Tensor:
                layer_w_in = w_in
            else:
                layer_w_in = None
            # end if

            # W bias
            if type(w_bias) is torch.Tensor and w_bias.ndim == 2:
                layer_w_bias = w_bias[n]
            elif type(w_bias) is torch.Tensor:
                layer_w_bias = w_bias
            else:
                layer_w_bias = None
            # end if

            # Parameters
            layer_sparsity = sparsity[n] if type(sparsity) is list else sparsity
            layer_input_set = input_set[n] if type(input_set) is list else input_set
            layer_w_sparsity = w_sparsity[n] if type(w_sparsity) is list else w_sparsity
            layer_nonlin_func = nonlin_func[n] if type(nonlin_func) is list else nonlin_func

            self.esn_layers.append(LiESNCell(
                layer_leaky_rate, False, layer_input_dim, hidden_dim[n], layer_spectral_radius, layer_bias_scaling,
                layer_input_scaling, layer_w, layer_w_in, layer_w_bias, None, layer_sparsity, layer_input_set,
                layer_w_sparsity, layer_nonlin_func
            ))
        # end for

        # Output layer
        self.output = RRCell(self.n_features, output_dim, ridge_param, False, with_bias, learning_algo)
    # end __init__

    ###############################################
    # PROPERTIES
    ###############################################

    # Hidden layer
    @property
    def hidden(self):
        """
        Hidden layer
        :return:
        """
        # Hidden states
        hidden_states = list()

        # For each ESN
        for esn_cell in self.esn_layers:
            hidden_states.append(esn_cell.hidden)
        # end for

        return hidden_states
    # end hidden

    # Hidden weight matrix
    @property
    def w(self):
        """
        Hidden weight matrix
        :return:
        """
        # W
        w_mtx = list()

        # For each ESN
        for esn_cell in self.esn_layers:
            w_mtx.append(esn_cell.w)
        # end for

        return w_mtx
    # end w

    # Input matrix
    @property
    def w_in(self):
        """
        Input matrix
        :return:
        """
        # W in
        win_mtx = list()

        # For each ESN
        for esn_cell in self.esn_layers:
            win_mtx.append(esn_cell.w_in)
        # end for

        return win_mtx
    # end w_in

    ###############################################
    # PUBLIC
    ###############################################

    # Reset learning
    def reset(self):
        """
        Reset learning
        :return:
        """
        self.output.reset()

        # Training mode again
        self.train(True)
    # end reset

    # Output matrix
    def get_w_out(self):
        """
        Output matrix
        :return:
        """
        return self.output.w_out
    # end get_w_out

    # Forward
    def forward(self, u, y=None):
        """
        Forward
        :param u: Input signal.
        :param y: Target outputs
        :return: Output or hidden states
        """
        # Hidden states
        hidden_states = Variable(torch.zeros(u.size(0), u.size(1), self.n_features))

        # Compute hidden states
        pos = 0
        for index, esn_cell in enumerate(self.esn_layers):
            layer_dim = esn_cell.output_dim
            if index == 0:
                last_hidden_states = esn_cell(u)
                hidden_states[:, :, pos:pos+layer_dim] = last_hidden_states
            else:
                last_hidden_states = esn_cell(last_hidden_states)
                hidden_states[:, :, pos:pos + layer_dim] = last_hidden_states
            # end if
            pos += layer_dim
        # end for

        # Learning algo
        return self.output(hidden_states, y)
    # end forward

    # Finish training
    def finalize(self):
        """
        Finalize training with LU factorization
        """
        # Finalize output training
        self.output.finalize()

        # Not in training mode anymore
        self.train(False)
    # end finalize

    # Reset hidden layer
    def reset_hidden(self):
        """
        Reset hidden layer
        :return:
        """
        self.esn_cell.reset_hidden()
    # end reset_hidden

    # Get W's spectral radius
    def get_spectral_radius(self):
        """
        Get W's spectral radius
        :return: W's spectral radius
        """
        return self.esn_cell.get_spectral_raduis()
    # end spectral_radius

    ############################################
    # STATIC
    ############################################

    # Generate W matrices for a stacked ESN
    @staticmethod
    def generate_ws(n_layers, reservoir_size, w_sparsity):
        """
        Generate W matrices for a stacked ESN
        :param n_layers:
        :param reservoir_size:
        :param w_sparsity:
        :return:
        """
        ws = torch.FloatTensor(n_layers, reservoir_size, reservoir_size)
        for i in range(n_layers):
            ws[i] = ESNCell.generate_w(reservoir_size, w_sparsity)
        # end for
        return ws
    # end for

# end ESNCell



# -*- coding: utf-8 -*-
#
# File : echotorch/nn/ESN.py
# Description : An Echo State Network module.
# Date : 26th of January, 2018
#
# This file is part of EchoTorch.  EchoTorch is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Nils Schaetti <nils.schaetti@unine.ch>

"""
Created on 26 January 2018
@author: Nils Schaetti
"""

# Imports
import torch.sparse
import torch
import torch.nn as nn
from torch.autograd import Variable


# Principal Component Analysis layer
class PCACell(nn.Module):
    """
    Principal Component Analysis layer
    """

    # Constructor
    def __init__(self, input_dim, output_dim):
        """
        Constructor
        :param input_dim: Inputs dimension.
        :param output_dim: Reservoir size
        """
        super(PCACell, self).__init__()
        pass
    # end __init__

    ###############################################
    # PROPERTIES
    ###############################################

    ###############################################
    # PUBLIC
    ###############################################

    # Reset learning
    def reset(self):
        """
        Reset learning
        :return:
        """
        # Training mode again
        self.train(True)
    # end reset

    # Forward
    def forward(self, x, y=None):
        """
        Forward
        :param x: Input signal.
        :param y: Target outputs
        :return: Output or hidden states
        """
        # Batch size
        batch_size = x.size()[0]

        # Time length
        time_length = x.size()[1]

        # Add bias
        if self.with_bias:
            x = self._add_constant(x)
        # end if
    # end forward

    # Finish training
    def finalize(self):
        """
        Finalize training with LU factorization or Pseudo-inverse
        """
        pass
    # end finalize

    ###############################################
    # PRIVATE
    ###############################################

    # Add constant
    def _add_constant(self, x):
        """
        Add constant
        :param x:
        :return:
        """
        bias = Variable(torch.ones((x.size()[0], x.size()[1], 1)), requires_grad=False)
        return torch.cat((bias, x), dim=2)
    # end _add_constant

# end PCACell



# -*- coding: utf-8 -*-
#
# File : echotorch/nn/ESNCell.py
# Description : An Echo State Network layer.
# Date : 26th of January, 2018
#
# This file is part of EchoTorch.  EchoTorch is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Nils Schaetti <nils.schaetti@unine.ch>

"""
Created on 26 January 2018
@author: Nils Schaetti
"""

import torch
import torch.sparse
from torch.autograd import Variable
import torch.nn as nn
import echotorch.utils
import numpy as np


# Echo State Network layer
class ESNCell(nn.Module):
    """
    Echo State Network layer
    """

    # Constructor
    def __init__(self, input_dim, output_dim, spectral_radius=0.9, bias_scaling=0, input_scaling=1.0, w=None, w_in=None,
                 w_bias=None, w_fdb=None, sparsity=None, input_set=[1.0, -1.0], w_sparsity=None,
                 nonlin_func=torch.tanh, feedbacks=False, feedbacks_dim=None, wfdb_sparsity=None,
                 normalize_feedbacks=False):
        """
        Constructor
        :param input_dim: Inputs dimension.
        :param output_dim: Reservoir size
        :param spectral_radius: Reservoir's spectral radius
        :param bias_scaling: Scaling of the bias, a constant input to each neuron (default: 0, no bias)
        :param input_scaling: Scaling of the input weight matrix, default 1.
        :param w: Internation weights matrix
        :param w_in: Input-reservoir weights matrix
        :param w_bias: Bias weights matrix
        :param sparsity:
        :param input_set:
        :param w_sparsity:
        :param nonlin_func: Reservoir's activation function (tanh, sig, relu)
        """
        super(ESNCell, self).__init__()

        # Params
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.spectral_radius = spectral_radius
        self.bias_scaling = bias_scaling
        self.input_scaling = input_scaling
        self.sparsity = sparsity
        self.input_set = input_set
        self.w_sparsity = w_sparsity
        self.nonlin_func = nonlin_func
        self.feedbacks = feedbacks
        self.feedbacks_dim = feedbacks_dim
        self.wfdb_sparsity = wfdb_sparsity
        self.normalize_feedbacks = normalize_feedbacks

        # Init hidden state
        self.register_buffer('hidden', self.init_hidden())

        # Initialize input weights
        self.register_buffer('w_in', self._generate_win(w_in))

        # Initialize reservoir weights randomly
        self.register_buffer('w', self._generate_w(w))

        # Initialize bias
        self.register_buffer('w_bias', self._generate_wbias(w_bias))

        # Initialize feedbacks weights randomly
        if feedbacks:
            self.register_buffer('w_fdb', self._generate_wfdb(w_fdb))
        # end if
    # end __init__

    ###############################################
    # PUBLIC
    ###############################################

    # Forward
    def forward(self, u, y=None, w_out=None):
        """
        Forward
        :param u: Input signal
        :param y: Target output signal for teacher forcing
        :param w_out: Output weights for teacher forcing
        :return: Resulting hidden states
        """
        # Time length
        time_length = int(u.size()[1])

        # Number of batches
        n_batches = int(u.size()[0])

        # Outputs
        outputs = Variable(torch.zeros(n_batches, time_length, self.output_dim))
        outputs = outputs.cuda() if self.hidden.is_cuda else outputs

        # For each batch
        for b in range(n_batches):
            # Reset hidden layer
            self.reset_hidden()

            # For each steps
            for t in range(time_length):
                # Current input
                ut = u[b, t]

                # Compute input layer
                u_win = self.w_in.mv(ut)

                # Apply W to x
                x_w = self.w.mv(self.hidden)

                # Feedback or not
                if self.feedbacks and self.training and y is not None:
                    # Current target
                    yt = y[b, t]

                    # Compute feedback layer
                    y_wfdb = self.w_fdb.mv(yt)

                    # Add everything
                    x = u_win + x_w + y_wfdb + self.w_bias
                elif self.feedbacks and not self.training and w_out is not None:
                    # Compute past output
                    yt = w_out.mv(self.hidden)

                    # Normalize
                    if self.normalize_feedbacks:
                        yt -= torch.min(yt)
                        yt /= torch.max(yt) - torch.min(yt)
                        yt /= torch.sum(yt)
                    # end if

                    # Compute feedback layer
                    y_wfdb = self.w_fdb.mv(yt)

                    # Add everything
                    x = u_win + x_w + y_wfdb + self.w_bias
                else:
                    # Add everything
                    x = u_win + x_w + self.w_bias
                # end if

                # Apply activation function
                x = self.nonlin_func(x)

                # Add to outputs
                self.hidden.data = x.view(self.output_dim).data

                # New last state
                outputs[b, t] = self.hidden
            # end for
        # end for

        return outputs
    # end forward

    # Init hidden layer
    def init_hidden(self):
        """
        Init hidden layer
        :return: Initiated hidden layer
        """
        return Variable(torch.zeros(self.output_dim), requires_grad=False)
        # return torch.zeros(self.output_dim)
    # end init_hidden

    # Reset hidden layer
    def reset_hidden(self):
        """
        Reset hidden layer
        :return:
        """
        self.hidden.fill_(0.0)
    # end reset_hidden

    # Get W's spectral radius
    def get_spectral_radius(self):
        """
        Get W's spectral radius
        :return: W's spectral radius
        """
        return echotorch.utils.spectral_radius(self.w)
    # end spectral_radius

    ###############################################
    # PRIVATE
    ###############################################

    # Generate W matrix
    def _generate_w(self, w):
        """
        Generate W matrix
        :return:
        """
        # Initialize reservoir weight matrix
        if w is None:
            w = self.generate_w(self.output_dim, self.w_sparsity)
        else:
            if callable(w):
                w = w(self.output_dim)
            # end if
        # end if

        # Scale it to spectral radius
        w *= self.spectral_radius / echotorch.utils.spectral_radius(w)

        return Variable(w, requires_grad=False)
    # end generate_W

    # Generate Win matrix
    def _generate_win(self, w_in):
        """
        Generate Win matrix
        :return:
        """
        # Initialize input weight matrix
        if w_in is None:
            if self.sparsity is None:
                w_in = self.input_scaling * (
                            np.random.randint(0, 2, (self.output_dim, self.input_dim)) * 2.0 - 1.0)
                w_in = torch.from_numpy(w_in.astype(np.float32))
            else:
                w_in = self.input_scaling * np.random.choice(np.append([0], self.input_set),
                                                             (self.output_dim, self.input_dim),
                                                             p=np.append([1.0 - self.sparsity],
                                                                         [self.sparsity / len(self.input_set)] * len(
                                                                             self.input_set)))
                w_in = torch.from_numpy(w_in.astype(np.float32))
            # end if
        else:
            if callable(w_in):
                w_in = w_in(self.output_dim, self.input_dim)
            # end if
        # end if

        return Variable(w_in, requires_grad=False)
    # end _generate_win

    # Generate Wbias matrix
    def _generate_wbias(self, w_bias):
        """
        Generate Wbias matrix
        :return:
        """
        # Initialize bias matrix
        if w_bias is None:
            w_bias = self.bias_scaling * (torch.rand(1, self.output_dim) * 2.0 - 1.0)
        else:
            if callable((w_bias)):
                w_bias = w_bias(self.output_dim)
            # end if
        # end if

        return Variable(w_bias, requires_grad=False)
    # end _generate_wbias

    # Generate Wfdb matrix
    def _generate_wfdb(self, w_fdb):
        """
        Generate Wfdb matrix
        :return:
        """
        # Initialize feedbacks weight matrix
        if w_fdb is None:
            if self.wfdb_sparsity is None:
                w_fdb = self.input_scaling * (
                        np.random.randint(0, 2, (self.output_dim, self.feedbacks_dim)) * 2.0 - 1.0)
                w_fdb = torch.from_numpy(w_fdb.astype(np.float32))
            else:
                w_fdb = self.input_scaling * np.random.choice(np.append([0], self.input_set),
                                                             (self.output_dim, self.feedbacks_dim),
                                                             p=np.append([1.0 - self.wfdb_sparsity],
                                                                         [self.wfdb_sparsity / len(
                                                                             self.input_set)] * len(
                                                                             self.input_set)))
                w_fdb = torch.from_numpy(w_fdb.astype(np.float32))
            # end if
        else:
            if callable(w_fdb):
                w_fdb = w_fdb(self.output_dim, self.feedbacks_dim)
            # end if
        # end if

        return Variable(w_fdb, requires_grad=False)
    # end _generate_wfdb

    ############################################
    # STATIC
    ############################################

    # Generate W matrix
    @staticmethod
    def generate_w(output_dim, w_sparsity=None):
        """
        Generate W matrix
        :param output_dim:
        :param w_sparsity:
        :return:
        """
        # Sparsity
        if w_sparsity is None:
            w = torch.rand(output_dim, output_dim) * 2.0 - 1.0
        else:
            w = np.random.choice([0.0, 1.0], (output_dim, output_dim),
                                 p=[1.0 - w_sparsity, w_sparsity])
            w[w == 1] = np.random.rand(len(w[w == 1])) * 2.0 - 1.0
            w = torch.from_numpy(w.astype(np.float32))

            # Return
            return w
        # end if
        return w
    # end generate_w

    # To sparse matrix
    @staticmethod
    def to_sparse(m):
        """
        To sparse matrix
        :param m:
        :return:
        """
        # Rows, columns and values
        rows = torch.LongTensor()
        columns = torch.LongTensor()
        values = torch.FloatTensor()

        # For each row
        for i in range(m.shape[0]):
            # For each column
            for j in range(m.shape[1]):
                if m[i, j] != 0.0:
                    rows = torch.cat((rows, torch.LongTensor([i])), dim=0)
                    columns = torch.cat((columns, torch.LongTensor([j])), dim=0)
                    values = torch.cat((values, torch.FloatTensor([m[i, j]])), dim=0)
                # end if
            # end for
        # end for

        # Indices
        indices = torch.cat((rows.unsqueeze(0), columns.unsqueeze(0)), dim=0)

        # To sparse
        return torch.sparse.FloatTensor(indices, values)
    # end to_sparse

# end ESNCell
